#ifndef MIE_SHA1_H_
#define MIE_SHA1_H_
/**
	@file md5.h
	@brief SHA1NX
	@author $Author: shigeo $
*/

#include <assert.h>
#include "../../../include.mie/mie_common.h"
#include "../../../include.mie/mie_util.h"
#include <string>
#include <sstream>
#include <iomanip>

namespace MIE {

class SHA1 {
private:
	uint32 K_[80];
	uint32 digest_[5];

	uint32 S(uint32 x, int s) const
	{
#ifdef _MSC_VER
		return _rotl(x, s);
#else
		return (x << s) | (x >> (32 - s));
#endif
	}

	uint32 f0(uint32 b, uint32 c, uint32 d) const { return (b & c) | (~b & d); }
	uint32 f1(uint32 b, uint32 c, uint32 d) const { return b ^ c ^ d; }
	uint32 f2(uint32 b, uint32 c, uint32 d) const { return (b & c) | (b & d) | (c & d); }
	uint32 f(int t, uint32 b, uint32 c, uint32 d) const
	{
		if (t < 20) {
			return f0(b, c, d);
		} else
		if (t < 40) {
			return f1(b, c, d);
		} else
		if (t < 60) {
			return f2(b, c, d);
		} else {
			return f1(b, c, d);
		}
	}

public:
	SHA1(const std::string& str)
	{
		static const uint32 tbl[] = {
			0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xca62c1d6
		};
		for (size_t i = 0; i < NUM_OF_ARRAY(tbl); i++) {
			for (int j = 0; j < 20; j++) {
				K_[i * 20 + j] = tbl[i];
			}
		}
		if (!str.empty()) {
			calc(&str[0], str.size());
		}
	}
	/**
		@param buf [in] obt@(64byte)
	*/
	void round(uint32 H[5], const uint8 *buf)
	{
		uint32 W[80];
		for (int i = 0; i < 16; i++) {
			W[i] = MIE::Util::get32bitAsBE(&buf[i * 4]);
		}
		for (int i = 16 ; i < 80; i++) {
			W[i] = S(W[i - 3] ^ W[i - 8] ^ W[i - 14] ^ W[i - 16], 1);
		}
		uint32 a = H[0];
		uint32 b = H[1];
		uint32 c = H[2];
		uint32 d = H[3];
		uint32 e = H[4];
		for (int i = 0; i < 80; i++) {
			uint32 tmp = S(a, 5) + f(i, b, c, d) + e + W[i] + K_[i];
			e = d;
			d = c;
			c = S(b, 30);
			b = a;
			a = tmp;
		}
		H[0] += a;
		H[1] += b;
		H[2] += c;
		H[3] += d;
		H[4] += e;
	}

	void calc(const void *_buf = 0, size_t bufSize = 0)
	{
		const uint8 *buf = reinterpret_cast<const uint8 *>(_buf);
		const uint64 totalBitSize = bufSize * 8;
		uint32 H[5] = {
			0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476, 0xc3d2e1f0
		};

		while (bufSize >= 64) {
			round(H, buf);
			buf += 64;
			bufSize -= 64;
		}
		/* c64byte */
		assert(bufSize < 64);

		uint8 last[64];
		memcpy(last, buf, bufSize);
		memset(&last[bufSize], 0, 64 - bufSize);
		last[bufSize] = static_cast<uint8>(0x80); /* top bit = 1 */
		if (bufSize >= 56) {
			round(H, last);
			memset(last, 0, sizeof(last));
		}
		MIE::Util::set32bitAsBE(uint32(totalBitSize >> 32), &last[56]);
		MIE::Util::set32bitAsBE(uint32(totalBitSize), &last[60]);
		round(H, last);

		for (size_t i = 0; i < NUM_OF_ARRAY(H); i++) {
			MIE::Util::set32bitAsBE(H[i], &digest_[i]);
		}
	}
	/**
		convert to printable string
	*/
	const std::string toString() const
	{
		std::ostringstream os;
		for (size_t i = 0; i < NUM_OF_ARRAY(digest_); i++) {
			os << std::hex << std::setw(8) << std::setfill('0') << MIE::Util::get32bitAsBE(&digest_[i]);
		}
		return os.str();
	}
};

} // end of namespace

#endif /* MIE_SHA1_H_ */
